/*
Copyright (C) 2004 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2004 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;

namespace SpeedTouch_Speed_Monitor
{
	/// <summary>
	/// Summary description for Form1.
	/// </summary>
	public class FormDisplay : System.Windows.Forms.Form
	{
        public Tools.GUI.Components.SimpleChart.CSimpleChart simple_chart=null;
        public Tools.GUI.Components.SimpleChart.CPlotInfoList plot_info_list_down=null;
        public Tools.GUI.Components.SimpleChart.CPlotInfoList plot_info_list_up=null;
		/// <summary>
		/// Required designer variable.
		/// </summary>
		private System.ComponentModel.Container components = null;

		public FormDisplay(ClassOptions obj_options)
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();
            this.ShowInTaskbar=false;
            this.simple_chart.Dock=DockStyle.Fill;
            // simple chart presentation style
            this.simple_chart.LegendPosition=Tools.Drawing.CDrawString.TEXT_POSITION.INSIDE_TOP_RIGHT;
            this.simple_chart.SpaceBetweenPlots=25;
            this.simple_chart.BarType=Tools.GUI.Components.SimpleChart.CSimpleChart.BAR_TYPE.MIXED;
            this.simple_chart.VerticalAxisNegativeValueSeemPositive=true;
            this.simple_chart.AutoTickFrequency=true;
            this.simple_chart.AutoSizeIncreaseOnly=false;
            this.simple_chart.VerticalLeftAxis.Label.Font=new Font("Arial",7);

            this.simple_chart.VerticalLeftAxis.Label.Label="kb/s";
            this.simple_chart.VerticalLeftAxis.Label.PositionRelativeToControl=Tools.Drawing.CDrawString.TEXT_POSITION.OUTSIDE_TOP_RIGHT;
            
            this.plot_info_list_down=this.simple_chart.AddPlotInfoList();
            // this.plot_info_list_down.Legend="Down";
            this.plot_info_list_up=this.simple_chart.AddPlotInfoList();
            // this.plot_info_list_up.Legend="Up";
            this.plot_info_list_down.FixedLength=20;
            this.plot_info_list_up.FixedLength=20;

            this.simple_chart.ShowLegend=false;

            this.reload_options(obj_options);

            this.simple_chart.Draw();
		}
        public void reload_options(ClassOptions obj_options)
        {
            this.TopMost=obj_options.display_topmost;
            this.Opacity=obj_options.opacity;
            if (obj_options.transparent_background)
                this.TransparencyKey=Color.FromArgb(obj_options.background_color);
            else
                this.TransparencyKey=Color.FromArgb(1);// hopping this color is not used 

            this.simple_chart.AutoRedraw=false;

            this.simple_chart.ShowHorizontalGrid=obj_options.show_grid;
            this.simple_chart.BackColor=System.Drawing.Color.FromArgb(obj_options.background_color);
            this.simple_chart.VerticalLeftAxis.Label.Brush=new SolidBrush(System.Drawing.Color.FromArgb(obj_options.text_color));
            this.simple_chart.HorizontalGridPen=new Pen(new SolidBrush(System.Drawing.Color.FromArgb(obj_options.grid_color)),1);
            this.simple_chart.HorizontalAxis.Pen=new Pen(new SolidBrush(System.Drawing.Color.FromArgb(obj_options.axis_color)),1);
            this.simple_chart.VerticalLeftAxis.Pen=new Pen(new SolidBrush(System.Drawing.Color.FromArgb(obj_options.axis_color)),1);
            this.simple_chart.HorizontalAxis.Pen.CustomEndCap=new System.Drawing.Drawing2D.AdjustableArrowCap(6, 6, false);
            this.simple_chart.VerticalLeftAxis.Pen.CustomEndCap=new System.Drawing.Drawing2D.AdjustableArrowCap(6, 6, false);
            this.simple_chart.GraphType=obj_options.graph_type;
            this.simple_chart.AutoTickFrequencyNumberOfTicks=obj_options.number_of_ticks;

            this.plot_info_list_down.GraphBrush=new SolidBrush(System.Drawing.Color.FromArgb(obj_options.download_color));
            this.plot_info_list_up.GraphBrush=new SolidBrush(System.Drawing.Color.FromArgb(obj_options.upload_color));

            this.simple_chart.AutoRedraw=true;
        }
        public void UpdateValue(float download,float upload)
        {
            this.Text="SpeedTouch Speed Monitor "+Math.Round(download,1).ToString()+"/"+Math.Round(upload,1).ToString()+" kbs";
            this.simple_chart.AutoRedraw=false;// to earn 1 refresh per Elapsed event
            this.plot_info_list_down.AddPlot(download);
            this.plot_info_list_up.AddPlot(-upload);
            this.simple_chart.AutoRedraw=true;
        }

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
            this.simple_chart = new Tools.GUI.Components.SimpleChart.CSimpleChart();
            this.SuspendLayout();
            // 
            // simple_chart
            // 
            this.simple_chart.AutoRedraw = true;
            this.simple_chart.AutoSize = true;
            this.simple_chart.AutoSizeIncreaseOnly = true;
            this.simple_chart.AutoTickFrequency = false;
            this.simple_chart.AutoTickFrequencyNumberOfTicks = 8;
            this.simple_chart.BarType = Tools.GUI.Components.SimpleChart.CSimpleChart.BAR_TYPE.SPLITTED;
            this.simple_chart.GraphType = Tools.GUI.Components.SimpleChart.CSimpleChart.GRAPH_TYPE.BAR;
            this.simple_chart.LabelsFont = new System.Drawing.Font("Arial", 10F);
            this.simple_chart.LegendFont = new System.Drawing.Font("Arial", 10F);
            this.simple_chart.LegendPosition = Tools.Drawing.CDrawString.TEXT_POSITION.OUTSIDE_TOP_RIGHT;
            this.simple_chart.Location = new System.Drawing.Point(0, 0);
            this.simple_chart.MaxValue = 50;
            this.simple_chart.MinValue = 0;
            this.simple_chart.Name = "simple_chart";
            this.simple_chart.ShowHorizontalGrid = false;
            this.simple_chart.ShowLegend = true;
            this.simple_chart.ShowPlotsLabel = true;
            this.simple_chart.ShowVerticalLeftAxisValues = true;
            this.simple_chart.ShowVerticalRightAxisValues = false;
            this.simple_chart.Size = new System.Drawing.Size(288, 78);
            this.simple_chart.SpaceBetweenPlots = 50;
            this.simple_chart.TabIndex = 0;
            this.simple_chart.VerticalAxisNegativeValueSeemPositive = false;
            this.simple_chart.VerticalAxisTickFrequency = 2F;
            this.simple_chart.VerticalLeftAxisValuesAngle = 0;
            this.simple_chart.VerticalRightAxisValuesAngle = 0;
            // 
            // FormDisplay
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(288, 78);
            this.Controls.Add(this.simple_chart);
            this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.SizableToolWindow;
            this.Name = "FormDisplay";
            this.StartPosition = System.Windows.Forms.FormStartPosition.Manual;
            this.Text = "SpeedTouch Speed Monitor";
            this.Closing += new System.ComponentModel.CancelEventHandler(this.FormDisplay_Closing);
            this.ResumeLayout(false);

        }
		#endregion

        private const int WM_QUERYENDSESSION=0x0011;
        protected override void WndProc(ref Message uMsg)
        {
            switch(uMsg.Msg)
            {
                case WM_QUERYENDSESSION:
                    this.force_close=true;
                    break;
            }
            base.WndProc(ref uMsg);
        }

        #region closing
        private bool force_close=false;
        public void Close()
        {
            this.force_close=true;
            base.Close();
        }

        private void FormDisplay_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            if (!this.force_close)
            {
                e.Cancel=true;
                this.Hide();
            }
        }
        #endregion

    }
}
